{
    "namespace": "suzieq",
    "name": "ospfIf",
    "type": "record",
    "fields": [
        {
	    "name": "sqvers",
	    "type": "string",
	    "partition": 0,
	    "default": "1.0"
	},
        {
            "name": "instance",
            "type": "string"
        },
        {
            "name": "cost",
            "type": "long"
        },
        {
            "name": "areaStub",
            "type": "boolean"
        },
        {
            "name": "retxTime",
            "type": "long"
        },
        {
            "name": "passive",
            "type": "boolean",
            "display": 6
        },
        {
            "name": "deadTime",
            "type": "long"
        },
        {
            "name": "helloTime",
            "type": "long"
        },
        {
            "name": "vrf",
            "type": "string",
            "key": 2,
            "display": 2
        },
        {
            "name": "area",
            "type": "string",
            "display": 4
        },
        {
            "name": "ifname",
            "type": "string",
            "key": 3,
            "display": 3
        },
	{
	    "name": "peerHostname",
	    "type": "string",
	    "display": 8,
	    "depends": "hostname vrf ipAddress"
        },
        {
            "name": "networkType",
            "type": "string"
        },
        {
            "name": "nbrCount",
            "type": "long",
            "display": 7
        },
        {
            "name": "maskLen",
            "type": "long"
        },
        {
            "name": "stubType",
            "type": "string"
        },
        {
            "name": "authType",
            "type": "string"
        },
        {
            "name": "noSummary",
            "type": "boolean"
        },
        {
            "name": "isBackbone",
            "type": "boolean"
        },
        {
            "name": "isUnnumbered",
            "type": "boolean"
        },
        {
            "name": "state",
            "type": "string",
            "display": 5
        },
        {
            "name": "ipAddress",
            "type": "string"
        },
        {
            "name": "routerId",
            "type": "string"
        },
        {
            "name": "hostname",
            "type": "string",
            "key": 1,
            "display": 1,
	    "partition": 2
        },
        {
            "name": "namespace",
            "type": "string",
            "key": 0,
            "display": 0,
	    "partition": 1
        },
        {
            "name": "timestamp",
            "type": "timestamp"
        },
        {
            "name": "active",
            "type": "boolean"
        }
    ]
}